#! /usr/bin/env python
#
# This script queries the Launchpad database for translation data and
# Returns the result in different formats
#
# Author: David Planella <david.planella@ubuntu.com>
#
# Copyright (c) 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import optparse
import os
import sys
import subprocess
import json
import gzip
import csv
import errno
from tempfile import TemporaryFile

# Program defaults. These act as a config file of some sort.
# Modify them in order to be able to call the script without
# specifying any options
HOSTNAME = 'sourcherry.canonical.com'
DATABASE = 'lpmain_staging'
PRIVATE_PATH = '/home/person/public_html/ubuntu'
PUBLIC_PATH = None # e.g. '/home/person/public_html/ubuntu'
SSH_KEY_PATH = None # e.g. '/home/person/.ssh/id_rookery_rsa'

# Executable paths of the auxiliary programs that need to be called
PSQL_BIN = '/usr/bin/psql'
SCP_BIN = '/usr/bin/scp'

# PostgreSQL queries to run against the given database to produce the desired
# output.
# If you need to add a new query, remember to add it to the QUERIES dict
# below
PACKAGES_QUERY = '''
select
    sourcepackagename.name AS sourcepackage,
    potemplate.translation_domain,
    potemplate.name AS template_name,
    potemplate.date_last_updated AS template_last_updated,
    pofile.date_changed AS translation_last_updated,
    language.code AS language,
    (pofile.currentcount+pofile.rosettacount) AS translated,
    pofile.unreviewed_count AS unreviewed,
    pofile.updatescount AS changed,
    potemplate.messagecount AS total
  from potemplate
  join sourcepackagename on sourcepackagename=sourcepackagename.id
  join distroseries on distroseries=distroseries.id
  join pofile on pofile.potemplate=potemplate.id
  join language on pofile.language=language.id
  where
    (potemplate.iscurrent or sourcepackagename.name='openoffice.org') and
    distroseries.name='{0}'
  order by language.code,
           sourcepackagename.name,
           potemplate.translation_domain,
           date_last_updated;'''

POTEMPLATES_QUERY = '''
select
    sourcepackagename.name AS sourcepackage,
    potemplate.translation_domain,
    potemplate.name AS template_name,
    potemplate.messagecount AS total,
    potemplate.iscurrent AS enabled,
    potemplate.languagepack,
    potemplate.priority,
    potemplate.date_last_updated
  from potemplate
  join sourcepackagename on sourcepackagename=sourcepackagename.id
  join distroseries on distroseries=distroseries.id
  where
    distroseries.name='{0}'
  order by sourcepackagename.name,
           potemplate.name;'''

# Add PostgreSQL queries here using the format
# 'output-filename': SQL_QUERY_VARIABLE
# Any query you add will be executed against the
# given database
QUERIES = {'package-stats': PACKAGES_QUERY,
           'potemplate-stats': POTEMPLATES_QUERY,
           }

class PsqlQuery(object):
    '''
    Executes PostgreSQL queries with psql and stores the results

    Supported storage formats for the results:
     - 'log': .log file, csv-like, delimited by the '|' character (used for
       compatibility with tools using this format already)
     - 'json': .json file containing a JSON structure

    '''

    def __init__(self, query, query_id, hostname, database, distro_codename,
                 private_path, public_path = None, ssh_key_path = None):
        self.query = query.format(distro_codename)
        self.query_id = query_id
        self.hostname = hostname
        self.database = database
        self.distro_codename = distro_codename
        self.private_path = private_path
        self.public_path = public_path
        self.ssh_key_path = ssh_key_path
        self.formats = ['log', 'json']

        # Check if the private path exists and try to create it if it
        # doesn't
        try:
            os.makedirs(self.private_path)
        except OSError, e:
            if not (e.errno == errno.EEXIST):
                logging.error(
                    "The {0} directory ".format(self.private_path) +
                    "does not exist and you don't seem to have " +
                    "permissions to create it")
                raise

        # Check if the public path exists and try to create it if it
        # doesn't
        if self.public_path:
            try:
                os.makedirs(self.public_path)
            except OSError, e:
                if not (e.errno == errno.EEXIST):
                    logging.error(
                        "The {0} directory ".format(self.public_path) +
                        "does not exist and you don't seem to have " +
                        "permissions to create it")
                    raise

        # Check if the psql executable exists
        if not (os.path.exists(PSQL_BIN) and os.access(PSQL_BIN, os.X_OK)):
            logging.error(
                "The {0} program ".format(PSQL_BIN) +
                "is missing on your system. You'll need to install the " +
                "'postgresql-client-common' to run this script")
            raise subprocess.CalledProcessError(errno.ENOENT,
                                                'psql binary check')

    def run(self):
        '''Run a PostgreSQL query through psql, with the given parameters'''

        # Original file where to store the results of the query
        orig_output_filename = \
            os.path.join(self.private_path,
            self.__get_output_filename('log'))

        # Run the query with the given parameters
        opts_args = [
            '-U', 'postgres',
            '-h', self.hostname,
            self.database,
            '-o', orig_output_filename,
            '-c', self.query,
        ]

        cmd = [PSQL_BIN] + opts_args
        logging.debug(
            'Running psql query - executing command...\n{0}'.format(
                ' '.join(cmd)))

        try:
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError:
            logging.error("Error executing the following " +
                          "command:\n{0}".format(' '.join(cmd)))
            raise

        # Publish the result in any of the given formats
        for format in self.formats:
            result = self.__postprocess_result(orig_output_filename, format)
            gzipfile = self.__archive_result(result)
            if self.public_path:
                self.__publish_result(gzipfile)

    def __get_output_filename(self, format):
        '''
        Builds and returns the full name of the output file, given its
        format
        '''
        return(self.distro_codename + '-' + self.query_id + '.' + format)

    def __postprocess_result(self, filename, format):
        '''
        Given the @filename containing the results of running a database
        query, in log format, postprocess it to generate other formats
        '''

        # We only support additional JSON output for now
        if format == 'json':
            # Remove all whitespace from the file
            with open(filename) as f:
                lines = [
                    line.translate(None, ' \t') for line in f.readlines()
                        if not line.startswith(('-', '('))
                ]
                in_string = ''.join(lines)
                logging.debug(in_string)
            # Create a temp file to be able to use the DictReader, and
            # generate a json file
            with TemporaryFile() as f:
                f.write(in_string)
                f.seek(0)
                reader = csv.DictReader(f, delimiter = '|')
                data = [row for row in reader]
                # Substitute string values by their real types
                for row in data:
                    for (k, value) in row.items():
                        if value.isdigit():
                            row[k] = int(value)
                        elif value == 't':
                            row[k] = True
                        elif value == 'f':
                            row[k] = False

                result_filename = os.path.splitext(filename)[0] + '.json'
                with open(result_filename, 'w') as fj:
                    json.dump(data, fj)
        elif format == 'log':
            result_filename = os.path.splitext(filename)[0] + '.log'

        return result_filename

    def __archive_result(self, filename):
        '''Creates a gzip archive of the given file and returns its name'''

        gzipfile = filename + '.gz'
        with open(filename, 'rb') as f_in:
            archive = gzip.open(gzipfile, 'wb')
            try:
                archive.writelines(f_in)
            finally:
                archive.close()

        return gzipfile

    def __publish_result(self, filename):
        '''
        Publishes the given file to a public location through scp.
        We're assuming that the host is at people.canonical.com.
        '''

        HOST = 'people.canonical.com'

        opts_args = [
                '-i', self.ssh_key_path,
                filename,
                HOST + ':' + self.public_path,
        ]
        cmd = [SCP_BIN] + opts_args
        logging.debug(
                'Publishing results - executing command...\n{0}'.format(
                    ' '.join(cmd)))
        try:
           subprocess.check_call(cmd)
        except subprocess.CalledProcessError:
            logging.error("Error executing the following " +
                          "command:\n{0}".format(' '.join(cmd)))
            raise

if __name__ == "__main__":

    usage = (
'''
%prog [options] distro_codename

Execute PostgreSQL queries to obtain a raw database dump in text files in
both csv and json formats, containing statistics about translation coverage
for all the languages and translatable packages under the Ubuntu project in
Launchpad.

As the primary output will most probably end up in an internal server, an
optional public server + ssh key can be specified for the script to copy
this output to the public server.

Examples:

--  Run the script having set all the defaults for options and specifying
    natty as the distro to get the data from:

    ./lp-get-ul10nstats.py natty

--  Run the script against a local launchpad instance for testing purposes,
    obtaining natty data:

    ./lp-get-ul10nstats.py -n localhost -d launchpad_dev natty
'''
        )
    parser = optparse.OptionParser(usage = usage)
    parser.add_option('-v', '--verbose', dest='verbose', action='store_true',
        help = 'Print debug info')
    parser.add_option("-n", "--hostname", action="store", dest="hostname",
                      help=('The host name of the machine on which the ' +
                            'database server is running.'))
    parser.add_option("-d", "--database", action="store", dest="database",
                      help="The name of the database to connect to.")
    parser.add_option("-p", "--private-path", action="store",
                      dest="private_path",
                      help=('The private path on the database server the ' +
                            'intermediate results will be stored on (e.g. ' +
                            '"/home/person/public_html/stats").'))
    parser.add_option("-u", "--public-path", action="store",
                      dest="public_path",
                      help=('The public path on the people.canonical.com ' +
                            'server where the final results will be ' +
                            'published (e.g. ' +
                            '"/home/person/public_html/stats").'))
    parser.add_option("-s", "--ssh-key-path", action="store",
                      dest="ssh_key_path",
                      help=('The path to the SSH key used for publishing ' +
                            'the results on the final public destination ' +
                            '(e.g. "/home/person/.ssh/id_rookery_rsa").'))
    parser.set_defaults(verbose = False,
                        hostname = HOSTNAME,
                        database = DATABASE,
                        private_path = PRIVATE_PATH,
                        public_path = PUBLIC_PATH,
                        ssh_key_path = SSH_KEY_PATH,
                        )
    (options, args) = parser.parse_args()

    if not args or len(args) > 1:
        logging.error(
            "You've either specified no arguments or too many of them.\n" +
            "You need to specify at least (and at most) a distro " +
            "codename (e.g. natty)")
        parser.print_help()
        sys.exit(1)

    distro_codename = args[0]

    # Set the verbosity
    logging_format = '%(asctime)s %(levelname)s %(message)s'
    if options.verbose:
        logging.basicConfig(level = logging.DEBUG,
                            format = logging_format)
    else:
        logging.basicConfig(format = logging_format)

    # Run all specified queries using psql
    for (query_id, query) in QUERIES.items():
        try:
            psql_query = PsqlQuery(query,
                                   query_id,
                                   options.hostname,
                                   options.database,
                                   distro_codename,
                                   options.private_path,
                                   options.public_path,
                                   options.ssh_key_path,
                                   )
            psql_query.run()
        except subprocess.CalledProcessError as err:
            logging.error(err)
            sys.exit(err.returncode)
